@@ -5,19 +5,20 @@ import (
 	"fmt"
 	"os"
 	"path/filepath"
+	"regexp"
 	"strings"
 	"time"
 
-	"github.com/zk-org/zk/internal/core"
-	"github.com/zk-org/zk/internal/util"
-	"github.com/zk-org/zk/internal/util/errors"
-	"github.com/zk-org/zk/internal/util/opt"
-	strutil "github.com/zk-org/zk/internal/util/strings"
 	"github.com/tliron/glsp"
 	protocol "github.com/tliron/glsp/protocol_3_16"
 	glspserv "github.com/tliron/glsp/server"
 	"github.com/tliron/kutil/logging"
 	_ "github.com/tliron/kutil/logging/simple"
+	"github.com/zk-org/zk/internal/core"
+	"github.com/zk-org/zk/internal/util"
+	"github.com/zk-org/zk/internal/util/errors"
+	"github.com/zk-org/zk/internal/util/opt"
+	strutil "github.com/zk-org/zk/internal/util/strings"
 )
 
 // Server holds the state of the Language Server.
@@ -396,10 +397,6 @@ func NewServer(opts ServerOpts) *Server {
 	}
 
 	handler.TextDocumentCodeAction = func(context *glsp.Context, params *protocol.CodeActionParams) (interface{}, error) {
-		if isRangeEmpty(params.Range) {
-			return nil, nil
-		}
-
 		doc, ok := server.documents.Get(params.TextDocument.URI)
 		if !ok {
 			return nil, nil
@@ -408,38 +405,44 @@ func NewServer(opts ServerOpts) *Server {
 
 		actions := []protocol.CodeAction{}
 
-		addAction := func(dir string, actionTitle string) error {
-			opts := cmdNewOpts{
-				Title: doc.ContentAtRange(params.Range),
-				Dir:   dir,
-				InsertLinkAtLocation: &protocol.Location{
-					URI:   params.TextDocument.URI,
-					Range: params.Range,
-				},
+		missingBacklinkActions := server.getMissingBacklinkCodeActions(doc, params.TextDocument.URI, params.Range)
+		actions = append(actions, missingBacklinkActions...)
+
+		// Only add "New note" actions if range is not empty.
+		if !isRangeEmpty(params.Range) {
+			addAction := func(dir string, actionTitle string) error {
+				opts := cmdNewOpts{
+					Title: doc.ContentAtRange(params.Range),
+					Dir:   dir,
+					InsertLinkAtLocation: &protocol.Location{
+						URI:   params.TextDocument.URI,
+						Range: params.Range,
+					},
+				}
+
+				var jsonOpts map[string]interface{}
+				err := unmarshalJSON(opts, &jsonOpts)
+				if err != nil {
+					return err
+				}
+
+				actions = append(actions, protocol.CodeAction{
+					Title: actionTitle,
+					Kind:  stringPtr(protocol.CodeActionKindRefactor),
+					Command: &protocol.Command{
+						Title:     actionTitle,
+						Command:   cmdNew,
+						Arguments: []interface{}{wd, jsonOpts},
+					},
+				})
+
+				return nil
 			}
 
-			var jsonOpts map[string]interface{}
-			err := unmarshalJSON(opts, &jsonOpts)
-			if err != nil {
-				return err
-			}
-
-			actions = append(actions, protocol.CodeAction{
-				Title: actionTitle,
-				Kind:  stringPtr(protocol.CodeActionKindRefactor),
-				Command: &protocol.Command{
-					Title:     actionTitle,
-					Command:   cmdNew,
-					Arguments: []interface{}{wd, jsonOpts},
-				},
-			})
-
-			return nil
+			addAction(wd, "New note in current directory")
+			addAction("", "New note in top directory")
 		}
 
-		addAction(wd, "New note in current directory")
-		addAction("", "New note in top directory")
-
 		return actions, nil
 	}
 
@@ -580,8 +583,7 @@ func (s *Server) refreshDiagnosticsOfDocument(doc *document, notify glsp.NotifyF
 	}
 
 	diagConfig := notebook.Config.LSP.Diagnostics
-	if diagConfig.WikiTitle == core.LSPDiagnosticNone && diagConfig.DeadLink == core.LSPDiagnosticNone {
-		// No diagnostic enabled.
+	if !diagConfig.IsEnabled() {
 		return
 	}
 
@@ -617,6 +619,12 @@ func (s *Server) refreshDiagnosticsOfDocument(doc *document, notify glsp.NotifyF
 				}
 				severity = protocol.DiagnosticSeverity(diagConfig.DeadLink)
 				message = "not found"
+			} else if target.URI == doc.URI {
+				if diagConfig.SelfLink == core.LSPDiagnosticNone {
+					continue
+				}
+				severity = protocol.DiagnosticSeverity(diagConfig.SelfLink)
+				message = "self-referential link"
 			} else {
 				if diagConfig.WikiTitle == core.LSPDiagnosticNone {
 					continue
@@ -633,6 +641,11 @@ func (s *Server) refreshDiagnosticsOfDocument(doc *document, notify glsp.NotifyF
 			})
 		}
 
+		if diagConfig.MissingBacklink.Level != core.LSPDiagnosticNone {
+			backlinks := s.getMissingBacklinkDiagnostics(doc, notebook, diagConfig.MissingBacklink)
+			diagnostics = append(diagnostics, backlinks...)
+		}
+
 		go notify(protocol.ServerTextDocumentPublishDiagnostics, protocol.PublishDiagnosticsParams{
 			URI:         doc.URI,
 			Diagnostics: diagnostics,
@@ -644,7 +657,11 @@ func (s *Server) refreshDiagnosticsOfDocument(doc *document, notify glsp.NotifyF
 // completion started automatically when typing an identifier, or manually.
 func (s *Server) buildInvokedCompletionList(notebook *core.Notebook, doc *document, position protocol.Position) ([]protocol.CompletionItem, error) {
 	currentWord := doc.WordAt(position)
-	if strings.HasPrefix(doc.LookBehind(position, len(currentWord)+2), "[[") {
+	// currentWord will include parentheses (( but not brackets [[.
+	// We check both if it's at len(currentWord) word and len(currentWord)-2 to account for auto-pair
+	if strings.HasPrefix(doc.LookBehind(position, len(currentWord)+2), "[[") ||
+		strings.HasPrefix(doc.LookBehind(position, len(currentWord)), "((") ||
+		(len(currentWord) >= 2 && strings.HasPrefix(doc.LookBehind(position, len(currentWord)-2), "((")) {
 		return s.buildLinkCompletionList(notebook, doc, position)
 	}
 
@@ -736,6 +753,12 @@ func (s *Server) buildLinkCompletionList(notebook *core.Notebook, doc *document,
 
 	var items []protocol.CompletionItem
 	for _, note := range notes {
+		// Exclude current note from completion results.
+		notePath := filepath.Join(notebook.Path, note.Path)
+		if notePath == doc.Path {
+			continue
+		}
+
 		item, err := s.newCompletionItem(notebook, note, doc, position, linkFormatter, templates)
 		if err != nil {
 			s.logger.Err(err)
@@ -811,7 +834,7 @@ func (s *Server) newCompletionItem(notebook *core.Notebook, note core.MinimalNot
 		addTextEdits := []protocol.TextEdit{}
 
 		startOffset := -2
-		if doc.LookBehind(pos, 2) != "[[" {
+		if (doc.LookBehind(pos, 2) != "[[") && (doc.LookBehind(pos, 2) != "((") {
 			currentWord := doc.WordAt(pos)
 			startOffset = -2 - len(currentWord)
 		}
@@ -848,11 +871,11 @@ func (s *Server) newTextEditForLink(notebook *core.Notebook, note core.MinimalNo
 	// Overwrite [[ trigger directly if the additional text edits are disabled.
 	startOffset := 0
 	if !s.useAdditionalTextEditsWithNotebook(notebook) {
-		if doc.LookBehind(pos, 2) == "[[" {
+		if (doc.LookBehind(pos, 2) == "[[") || (doc.LookBehind(pos, 2) == "((") {
 			startOffset = -2
 		} else {
 			currentWord := doc.WordAt(pos)
-			startOffset = -2 - len(currentWord) 
+			startOffset = -2 - len(currentWord)
 		}
 	}
 
@@ -934,3 +957,279 @@ func toBool(obj interface{}) bool {
 	s := strings.ToLower(fmt.Sprint(obj))
 	return s == "true" || s == "1"
 }
+
+// MissingBacklink contains information of a note that links to the current note.
+type MissingBacklink struct {
+	SourcePath  string
+	SourceTitle string
+}
+
+// findMissingBacklinks finds notes that link to the current note but are not linked back.
+func (s *Server) findMissingBacklinks(currentNoteID core.NoteID, notebook *core.Notebook, doc *document) ([]MissingBacklink, error) {
+	currentNote, err := notebook.FindMinimalNotes(core.NoteFindOpts{
+		IncludeIDs: []core.NoteID{currentNoteID},
+	})
+	if err != nil || len(currentNote) == 0 {
+		return nil, err
+	}
+	currentNotePath := currentNote[0].Path
+
+	notesThatLinkToUs, err := notebook.FindMinimalNotes(core.NoteFindOpts{
+		LinkTo: &core.LinkFilter{
+			Hrefs: []string{currentNotePath},
+		},
+	})
+	if err != nil {
+		return nil, err
+	}
+
+	if len(notesThatLinkToUs) == 0 {
+		return nil, nil
+	}
+
+	currentDocLinks, err := doc.DocumentLinks()
+	if err != nil {
+		return nil, err
+	}
+
+	backlinkedNoteIDs := make(map[core.NoteID]bool)
+	for _, link := range currentDocLinks {
+		if strutil.IsURL(link.Href) {
+			continue
+		}
+		// Resolve the link to find the target note
+		target, err := s.noteForLink(link, notebook)
+		if err != nil || target == nil {
+			continue
+		}
+		// Extract note ID from the URI - we need to find the note by path
+		targetPath, err := uriToPath(target.URI)
+		if err != nil {
+			continue
+		}
+		targetRelPath, err := notebook.RelPath(targetPath)
+		if err != nil {
+			continue
+		}
+		targetNote, err := notebook.FindByHref(targetRelPath, false)
+		if err != nil || targetNote == nil {
+			continue
+		}
+		backlinkedNoteIDs[targetNote.ID] = true
+	}
+
+	var missingBacklinks []MissingBacklink
+	for _, linkingNote := range notesThatLinkToUs {
+		if !backlinkedNoteIDs[linkingNote.ID] {
+			missingBacklinks = append(missingBacklinks, MissingBacklink{
+				SourcePath:  linkingNote.Path,
+				SourceTitle: linkingNote.Title,
+			})
+		}
+	}
+
+	return missingBacklinks, nil
+}
+
+// getMissingBacklinkDiagnostics returns diagnostics for missing backlinks in the current document.
+func (s *Server) getMissingBacklinkDiagnostics(doc *document, notebook *core.Notebook, config core.MissingBacklinkConfig) []protocol.Diagnostic {
+	relPath, err := notebook.RelPath(doc.Path)
+	if err != nil {
+		s.logger.Err(err)
+		return nil
+	}
+
+	currentNote, err := notebook.FindByHref(relPath, false)
+	if err != nil {
+		s.logger.Err(err)
+		return nil
+	}
+
+	if currentNote == nil {
+		return nil
+	}
+
+	missingBacklinks, err := s.findMissingBacklinks(currentNote.ID, notebook, doc)
+	if err != nil {
+		s.logger.Err(err)
+		return nil
+	}
+
+	lines := strings.Split(doc.Content, "\n")
+
+	var targetLine uint32
+	switch config.Position {
+	case core.LSPDiagnosticPositionTop:
+		targetLine = 0
+	case core.LSPDiagnosticPositionBottom:
+		targetLine = uint32(len(lines) - 1)
+	case core.LSPDiagnosticPositionLastSection:
+		targetLine = findLastSectionLine(lines)
+	default:
+		targetLine = uint32(len(lines) - 1) // Fallback to bottom
+	}
+
+	var diagnostics []protocol.Diagnostic
+	for _, backlink := range missingBacklinks {
+		diagSeverity := protocol.DiagnosticSeverity(config.Level)
+
+		noteTitle := backlink.SourceTitle
+		if noteTitle == "" {
+			noteTitle = backlink.SourcePath
+		}
+
+		// Convert path to be relative to current note's directory
+		currentNoteDir := filepath.Dir(doc.Path)
+		sourceNotePath := filepath.Join(notebook.Path, backlink.SourcePath)
+		relativePath, err := filepath.Rel(currentNoteDir, sourceNotePath)
+		if err != nil {
+			// Fallback to original path if relative path calculation fails
+			relativePath = backlink.SourcePath
+		}
+
+		message := fmt.Sprintf("Missing backlink to [%s](%s)", noteTitle, relativePath)
+
+		diagnostics = append(diagnostics, protocol.Diagnostic{
+			Range: protocol.Range{
+				Start: protocol.Position{Line: targetLine, Character: 0},
+				End:   protocol.Position{Line: targetLine, Character: 0},
+			},
+			Severity: &diagSeverity,
+			Source:   stringPtr("zk"),
+			Message:  message,
+		})
+	}
+
+	return diagnostics
+}
+
+// findLastSectionLine finds the line number of the last heading in the document.
+// Falls back to line 0 (top) if no headings are found.
+func findLastSectionLine(lines []string) uint32 {
+	headingRegex := regexp.MustCompile(`^#{1,6}\s+`)
+	var lastHeadingLine uint32 = 0 // Fallback to top
+	for i, line := range lines {
+		if headingRegex.MatchString(line) {
+			lastHeadingLine = uint32(i)
+		}
+	}
+	return lastHeadingLine
+}
+
+// getMissingBacklinkCodeActions returns code actions for adding missing backlinks.
+func (s *Server) getMissingBacklinkCodeActions(doc *document, docURI protocol.DocumentUri, requestRange protocol.Range) []protocol.CodeAction {
+	notebook, err := s.notebookOf(doc)
+	if err != nil {
+		return nil
+	}
+
+	diagConfig := notebook.Config.LSP.Diagnostics
+	if diagConfig.MissingBacklink.Level == core.LSPDiagnosticNone {
+		return nil
+	}
+
+	relPath, err := notebook.RelPath(doc.Path)
+	if err != nil {
+		return nil
+	}
+
+	currentNote, err := notebook.FindByHref(relPath, false)
+	if err != nil || currentNote == nil {
+		return nil
+	}
+
+	missingBacklinks, err := s.findMissingBacklinks(currentNote.ID, notebook, doc)
+	if err != nil || len(missingBacklinks) == 0 {
+		return nil
+	}
+
+	var actions []protocol.CodeAction
+	var formattedLinks []string
+
+	lines := strings.Split(doc.Content, "\n")
+	currentLine := min(requestRange.End.Line, uint32(len(lines)-1))
+
+	insertPosition := protocol.Position{
+		Line:      currentLine,
+		Character: uint32(len(lines[currentLine])),
+	}
+
+	// Format all links once and generate individual actions
+	for _, backlink := range missingBacklinks {
+		// Full note metadata is required for NewLinkFormatterContext
+		sourceNote, err := notebook.FindByHref(backlink.SourcePath, false)
+		if err != nil || sourceNote == nil {
+			continue
+		}
+
+		linkFormatter, err := notebook.NewLinkFormatter()
+		if err != nil {
+			continue
+		}
+
+		noteDir := filepath.Dir(doc.Path)
+		linkPath := core.NotebookPath{
+			Path:       backlink.SourcePath,
+			BasePath:   notebook.Path,
+			WorkingDir: noteDir,
+		}
+		linkContext, err := core.NewLinkFormatterContext(linkPath, backlink.SourceTitle, sourceNote.Metadata)
+		if err != nil {
+			continue
+		}
+
+		link, err := linkFormatter(linkContext)
+		if err != nil {
+			continue
+		}
+
+		// Store formatted link for reuse
+		formattedLinks = append(formattedLinks, link)
+
+		title := backlink.SourceTitle
+		if title == "" {
+			title = filepath.Base(backlink.SourcePath)
+		}
+
+		// Create individual action
+		actions = append(actions, protocol.CodeAction{
+			Title: fmt.Sprintf("Add backlink to %s", title),
+			Kind:  stringPtr(protocol.CodeActionKindQuickFix),
+			Edit: &protocol.WorkspaceEdit{
+				Changes: map[protocol.DocumentUri][]protocol.TextEdit{
+					docURI: {{
+						Range: protocol.Range{
+							Start: insertPosition,
+							End:   insertPosition,
+						},
+						NewText: "\n" + link,
+					}},
+				},
+			},
+		})
+	}
+
+	// Add "add all missing backlinks" action if there are multiple backlinks
+	if len(formattedLinks) > 1 {
+		// Join all formatted links with newlines
+		allLinksText := "\n" + strings.Join(formattedLinks, "\n")
+
+		actions = append(actions, protocol.CodeAction{
+			Title: fmt.Sprintf("Add all %d missing backlinks", len(formattedLinks)),
+			Kind:  stringPtr(protocol.CodeActionKindQuickFix),
+			Edit: &protocol.WorkspaceEdit{
+				Changes: map[protocol.DocumentUri][]protocol.TextEdit{
+					docURI: {{
+						Range: protocol.Range{
+							Start: insertPosition,
+							End:   insertPosition,
+						},
+						NewText: allLinksText,
+					}},
+				},
+			},
+		})
+	}
+
+	return actions
+}
